#ifndef lint
static char SCCSid[] = "@(#) ./comm/io/piofb.c 12/01/93";
#endif

/* 
   This file contains the routines for writing fixed blocks in parallel.
   We assume that each task has access to separate disk system; the writes
   and reads need not be synhronized.

   The algorithm is to allocate and manage a read and a write cache into
   which data may be copied before being written.  There are lots of 
   assumptions here about the lack of alaising of the contents of the file; 
   if another method is used to access the file, bad things may happen.

   In order to improve performance, if possible, data is read or written 
   directly from/to the device.  For more generality, variable sized data
   may be used.

   These routines are also structured so that they may be used by sequential
   applications.

   In addition, these routines are prepared to make use of non-blocking
   IO when available.

   An additional optimization that can be taken is to identify a block of
   read/write operations so that random positionings within a file 

   Early trials showed significant performance improvements for unformatted
   writes (factor of 2 on Suns).

   This interface DOES give us the ability to use non-blocking IO routines 
   since we can manage the read and write buffers directly.  Such an interface
   will need multiple cache buffers so that one buffer can drain or fill as 
   the other is used.

   The same kind of thing can be done in the parallel routines to 
   cache up the contributing messages.
 */

#include <stdio.h>
#include "tools.h"
#include "comm/comm.h"
#include "comm/io/pio.h"

#ifndef WRITE
#define WRITE(fd,b,n) while (1) {int err;\
     err = write(fd,b,n); if (err==-1 && errno == EINTR)continue;}
#endif

PIFcache *PIFCacheCreate( rsize, wsize )
int rsize, wsize;
{
PIFcache *new;
int      n;
char     *p;

new = NEW(PIFcache);    CHKPTRN(new);
if (rsize < 0) rsize = PIF_RCACHE_SIZE;
if (wsize < 0) wsize = PIF_WCACHE_SIZE;
new->rlen = rsize;
new->rsize= -1;
new->wlen = wsize;
new->rp   = 0;
new->wp   = 0;
new->rb   = 0;
new->wb   = 0;
new->rrloc= 0;
new->wrloc= 0;
new->fpos = 0;
n         = rsize + wsize;
p         = (char *)MALLOC( n );    CHKPTRN(p);
if (rsize) {
    new->rb = p; 
    p       += rsize;
    }
if (wsize) {
    new->wb = p;
    }
return new;
}

void PIFCacheDestroy( c )
PIFcache *c;
{
if (c->rlen) {
    FREE(c->rb);
    }
else {
    FREE(c->wb);
    }
FREE( c );
}

int PIFiWriteCached( fp, buffer, len )
PIFILE   *fp;
char     *buffer;
int      len;
{
int      m, n, nb, lenout;
PIFcache *c = fp->c;

lenout = 0;
/* Fill up cache; flush it if we fill it up */
while (len > 0 && (c->wp > 0 || len < c->wlen)) {
    m = len;
    if (len > c->wlen - c->wp) m = c->wlen - c->wp;
    memcpy( c->wb + c->wp, buffer, m );
    len    -= m;
    buffer += m;
    c->wp  += m;
    lenout += m;
    if (c->wp == c->wlen) {
	/* Flush cache */
	n        = PIiDefWrite( fp, c->wb, c->wlen );
	c->fpos  += n;
	c->wrloc = c->fpos;
	c->wp    = 0;
	}
    }
if (len > 0) {
    /* If we got here, we can write a multiple of blocksize out */
    nb     = len / c->wlen;
    n      = PIiDefWrite( fp, buffer, nb * c->wlen );
    buffer += n;
    len    -= n;
    lenout += n;
    c->fpos += n;
    c->wrloc = c->fpos;
    /* Place remainder into cache */
    if (len > 0) {
	memcpy( c->wb, buffer, len );
	c->wp += len;
	lenout += len;
	}
    }
return lenout;
}

int PIFiReadCached( fp, buffer, len )
PIFILE   *fp;
char     *buffer;
int      len;
{
int      m, n, nb, lenin;
PIFcache *c = fp->c;

lenin  = 0;
/* Get data from cache */
if (c->rsize > 0) {
    m = len;
    if (len > c->rsize - c->rp) m = c->rsize - c->rp;
    memcpy( buffer, c->rb + c->rp, m );
    len    -= m;
    buffer += m;
    c->rp  += m;
    lenin  += m;
    }
if (len > 0) {
    /* We can read some blocks directly into memory (using a multiple of
       blocksize) */
    nb     = len / c->rlen;
    if (nb > 0) {
	n        = PIiDefRead( fp, buffer, nb * c->wlen );
	buffer   += n;
	len      -= n;
	lenin    += n;
	c->rrloc = c->fpos;
	c->fpos  += n;
	}
    /* Get the remainder by loading the cache */
    if (len > 0) {
	/* This may read too much; needs to handle EOF (PIiDefRead 
	   fails for n < c->rlen; not the desired behavior) */
	n        = PIiDefRead( fp, c->rb, c->rlen );
	c->rsize = n;
	c->rp    = 0;
	c->rrloc = c->fpos;
	c->fpos  += n;
	if (n < len) len = n;
	memcpy( buffer, c->rb, len );
	c->rp    += len;
	lenin    += len;
	}
    }
return lenin;
}

int PIFiFlushCache( fp )
PIFILE *fp;
{
PIFcache *c = fp->c;
int      n;

if (c->wp > 0) {
    /* use c->wlen when writing fixed sized blocks */
    n     = PIiDefWrite( fp, c->wb, c->wp );
    fp->fpos += n;
    c->wrloc = fp->fpos;
    c->wp = 0;
    }
}

#ifndef SEEK_SET
#include <unistd.h>
#endif
/* For now, a seek discards any cached data.  This is very suboptimal; 
   At the very least, we should permit repositioning within the cache block. 
   This is a problem for writes unless the cache size exactly matches the
   IO block size (otherwise, we need to read to fill the cache).  We 
   will start by checking for the cases where we are seeking to the current
   location */
int PIFiLseekCached( fp, offset, whence, rw )
PIFILE *fp;
int    offset;
int    whence;
int    rw;         /* 1 for write, 0 for read */
{
PIFcache *c = fp->c;

if (whence == SEEK_SET) {
    if (rw && offset == c->wrloc + c->wp) return offset;
    /* 
    fprintf( stdout, "moving seek pointer to %d from %d\n", 
	     offset, c->wrloc + c->wp ); 
     */
    /* Flush the write cache */
    PIFiFlushCache( fp );
    /* If we are seeking within the read cache, just move the rp */
    if (!rw) {
	if (c->rsize > 0 && offset >= c->rrloc && 
	    c->rrloc + c->rsize > offset) {
	    c->rp = offset - c->rrloc;
	    return offset;
	    }
	fprintf( stdout, "moving seek pointer to %d from %d\n", 
		 offset, c->rrloc + c->rp ); 
	}
    return PIiDefLseek( fp, offset, whence, rw );
    }
else {
    /* Try to move the cache pointers */
    PIFiFlushCache( fp );
    return PIiDefLseek( fp, offset, whence, rw );
    }
}

/*@
  PIFCacheEnable - Enables the read/write caches

  Input Parameters:
. fp - PIF file pointer
. rsize - read cache size
. wsize - write cache size

  Notes:
  This routine enables the caching of data for better IO performance.  If 
  a file is being accessed sequentially (a common case), then these routines
  may significantly improve performance.
 @*/
void PIFCacheEnable( fp, rsize, wsize )
PIFILE *fp;
int    rsize, wsize;
{
if (fp->pmode == PIO_AS_PARALLEL ||
    (fp->pmode == PIO_AS_SEQUENTIAL && PSISROOT(fp->procset))) {
    fp->c = PIFCacheCreate( rsize, wsize );
    /* Change to cached readers and writers */
    if (wsize != 0) 
	fp->write = PIFiWriteCached;
    if (rsize != 0)
	fp->read  = PIFiReadCached;
    fp->flush = PIFiFlushCache;
    fp->lseek = PIFiLseekCached;
    }
}

